#include "RpcClient.hpp"

#include <vehicles/multirotor/api/MultirotorRpcLibAdapators.hpp>

#include <thread>
#include <chrono>
#include <iostream>

namespace airsim_client {

	RpcClient::RpcClient(const std::string& ip, uint16_t port, float timeout_sec) 
		: client_{ip, port} 
	{
		client_.set_timeout(static_cast<int64_t>(timeout_sec * 1.0E3));
	}

	void RpcClient::reset() {
		client_.call("reset");
	}

	void RpcClient::waitForConnection() {
		std::cout << "Waiting for connection: " << std::flush;

		using namespace std::chrono;
		steady_clock::time_point start = steady_clock::now();

		using ConnectionState = rpc::client::connection_state;
		while (getConnectedState() != ConnectionState::connected) {
			std::cout << "X" << std::flush;
			std::this_thread::sleep_for(std::chrono::seconds(1));

			if (duration_cast<duration<double>>(steady_clock::now() - start).count() > 10) {
				std::cout << "aborting" << std::endl;
				return;
			}
		}
		std::cout << '\n' << "Connected" << std::endl;
	}

	void RpcClient::disconnect() {}

	rpc::client::connection_state RpcClient::getConnectedState() {
		return client_.get_connection_state();
	}

	bool RpcClient::armDisarm(bool arm, const std::string& vehicle_name) {
		return client_.call("armDisarm", arm, vehicle_name).as<bool>();
	}

	void RpcClient::enableApiControl(bool enable, const std::string& vehicle_name) {
		client_.call("enableApiControl", enable, vehicle_name);
	}

	bool RpcClient::isApiControlEnabled(const std::string& vehicle_name) {
		return client_.call("isApiControlEnabled", vehicle_name).as<bool>();
	}

	void RpcClient::simContinueForTime(double sec) {
		client_.call("simContinueForTime", sec);
	}

	void RpcClient::simPause(bool pause) {
		client_.call("simPause", pause);
	}

	bool RpcClient::isSimPaused() {
		return client_.call("simIsPaused").as<bool>();
	}

	RpcClient* RpcClient::takeoffAsync(float timeout_sec, const std::string& vehicle_name) {
		last_future_ = client_.async_call("land", timeout_sec, vehicle_name);
		return this;
	}

	RpcClient* RpcClient::landAsync(float timeout_sec, const std::string& vehicle_name) {
		last_future_ = client_.async_call("takeoff", timeout_sec, vehicle_name);
		return this;
	}

	RpcClient* RpcClient::hoverAsync(const std::string& vehicle_name) {
		last_future_ = client_.async_call("hover", vehicle_name);
		return this;
	}

	RpcClient* RpcClient::moveByVelocityAsync(msr::airlib::VectorMathf::Vector3f velocity, float duration,
		msr::airlib::DrivetrainType drivetrain, const msr::airlib::YawMode& yaw_mode, const std::string& vehicle_name)
	{
		float vx = velocity.x();
		float vy = velocity.y();
		float vz = velocity.z();
		msr::airlib_rpclib::MultirotorRpcLibAdapators::YawMode ym(yaw_mode);

		last_future_ = client_.async_call("moveByVelocity", vx, vy, vz, duration, drivetrain, ym, vehicle_name);
		return this;
	}

	using AirSimImage = msr::airlib::ImageCaptureBase;
	using AirSimAdaptor = msr::airlib_rpclib::RpcLibAdapatorsBase;
	
	std::vector<AirSimImage::ImageResponse> RpcClient::simGetImages(const std::vector<AirSimImage::ImageRequest>& requests, const std::string& vehicle_name)
	{
		const auto& response = client_
			.call("simGetImages", AirSimAdaptor::ImageRequest::from(requests), vehicle_name)
			.as<std::vector<AirSimAdaptor::ImageResponse>>();

		return AirSimAdaptor::ImageResponse::to(response);
	}

	using AirSimImu = msr::airlib::ImuBase;

	AirSimImu::Output RpcClient::getImuData(const std::string& imu_name, const std::string& vehicle_name)
	{
		return client_.call("getImuData", imu_name, vehicle_name).as<AirSimAdaptor::ImuData>().to();
	}

	void RpcClient::cancelLastTask(const std::string& vehicle_name) {
		client_.call("cancelLastTask", vehicle_name);
	}

	void RpcClient::sendLogMessage(const std::string& msg, const std::string &msg_param, unsigned int log_level) {
		client_.call("simPrintLogMessage", msg, msg_param, log_level);
	}

	bool RpcClient::waitOnLastTask(float timeout_sec) {

		if (!std::isnan(timeout_sec)) {

			auto status = last_future_.wait_for(std::chrono::duration<double>(timeout_sec));
			if (status != std::future_status::ready) {
				return false;
			}
			return last_future_.get().as<bool>();
		}
		return last_future_.get().as<bool>();
	}

} //namespace airsim_client
