#include "Vehicle.hpp"

namespace airsim_client {

	using Vec3r = msr::airlib::Vector3r;
	using Rotation = msr::airlib::AirSimSettings::Rotation;
	using VehicleSetting = msr::airlib::AirSimSettings::VehicleSetting;
	using CameraSetting = msr::airlib::AirSimSettings::CameraSetting;
	using CaptureSetting = msr::airlib::AirSimSettings::CaptureSetting;

	Vehicle::Vehicle(std::string map_frame_name, VehicleSetting& settings)
		: 	map_frame_name_{std::move(map_frame_name)}, vehicle_name_{settings.vehicle_name}
	{
		convertNanToZeroInPose(settings.position, settings.rotation);
		addSpawnPose(settings);

		for (auto& camera : settings.cameras) {
			auto& camera_settings = camera.second;
			convertNanToZeroInPose(camera_settings.position, camera_settings.rotation);
			addCameraTransform(camera.first, camera_settings);

			for (auto& capture_setting : camera_settings.capture_settings) {
				if (!std::isnan(capture_setting.second.fov_degrees)) {
					addCameraInfo(camera.first, capture_setting.second);
				}
			}
		}

		//Todo sensor transforms
	}

	using Kinematics = msr::airlib::Kinematics;
	void Vehicle::updateCurrentTransform(const Kinematics::State& kinematics, TimeStamp time_stamp) {

		current_transform_ = Transform{
			map_frame_name_,
			vehicle_name_ + "/body",
			time_stamp,
			kinematics.pose
		};
	}

	void Vehicle::addSpawnPose(const VehicleSetting& settings) {
		using VecMath = msr::airlib::VectorMathf;
		using Utils = msr::airlib::Utils;

		Transform tf;

		tf.parent_frame_id = map_frame_name_;
		tf.child_frame_id = vehicle_name_ + "/spawn_pose";

		tf.transform.orientation = VecMath::toQuaternion(
			Utils::degreesToRadians(settings.rotation.pitch),
			Utils::degreesToRadians(settings.rotation.roll),
			Utils::degreesToRadians(settings.rotation.yaw));

		tf.transform.position = settings.position;

		spawn_pose_ = tf;
		static_transforms_.push_back(tf);
	}

	void Vehicle::addCameraTransform(const std::string& camera_name, const CameraSetting& settings) {

		//Rotation between world and body
		Transform tf_ned;
		tf_ned.parent_frame_id = vehicle_name_ + "/body";
		tf_ned.child_frame_id = tf_ned.parent_frame_id + "/" + camera_name;

		using Utils = msr::airlib::Utils;
		tf_ned.transform.orientation = msr::airlib::VectorMathf::toQuaternion(
			Utils::degreesToRadians(settings.rotation.pitch),
			Utils::degreesToRadians(settings.rotation.roll),
			Utils::degreesToRadians(settings.rotation.yaw));

		tf_ned.transform.position = settings.position;
		static_transforms_.push_back(tf_ned);

		//Rotation between body and optical
		Transform tf_optical;
		tf_optical.parent_frame_id = tf_ned.child_frame_id;
		tf_optical.child_frame_id = tf_optical.parent_frame_id + "/optical";

		tf_optical.transform.orientation = msr::airlib::VectorMathf::toQuaternion(0, M_PI/2, M_PI/2);
		static_transforms_.push_back(tf_optical);
	}

	void Vehicle::addCameraInfo(const std::string& camera_name, const CaptureSetting& capture_setting) {
		CameraInfo info;
		info.image_type = capture_setting.image_type;
		info.optical_frame_id = vehicle_name_ + "/body/" + camera_name + "/optical";
		info.image_width = capture_setting.width;
		info.image_height = capture_setting.height;

		double fx = (capture_setting.width/2.0)/std::tan(msr::airlib::Utils::degreesToRadians(capture_setting.fov_degrees / 2.0));
		double fy = (capture_setting.height/2.0)/std::tan(msr::airlib::Utils::degreesToRadians(capture_setting.fov_degrees / 2.0));

		info.intrinsic_matrix <<
			fx, 	0.0, 	capture_setting.width/2.0,
			0.0,	fy,		capture_setting.height/2.0,
			0.0,	0.0,	1.0;

		info.position = msr::airlib::VectorMathf::Vector3f::Zero();

		camera_infos_[info.image_type] = info;
	}

	void Vehicle::convertNanToZeroInPose(Vec3r& position, Rotation& rotation) {

		if (std::isnan(position.x())) {
			position.x() = 0.0;
		}
		if (std::isnan(position.y())) {
			position.y() = 0.0;
		}
		if (std::isnan(position.z())) {
			position.z() = 0.0;
		}
		if (std::isnan(rotation.yaw)) {
			rotation.yaw = 0.0;
		}
		if (std::isnan(rotation.pitch)) {
			rotation.pitch = 0.0;
		}
		if (std::isnan(rotation.roll)) {
			rotation.roll = 0.0;
		}
	}
} //namespace airsim_client
