#include "Client.hpp"
#include "Log.hpp"
#include "Input.hpp"
//#include "Nodes/AirSim/AirSimNode.hpp"
//#include "Nodes/Ros/RosNode.hpp"

namespace airsim_client { namespace core {

	GLFWwindow* Input::s_window_ = nullptr;

	using clock = std::chrono::steady_clock;

	Client::Client() {
		Log::Init();
		CLIENT_INFO("Logger initialized.");

		window_ = std::make_unique<Window>();
		window_->SetEventCallback([this](Event& e) {
			OnEvent(e);
		});

		Input::SetInputWindow(window_->GetGLFWwindowPtr());
	}

	void Client::AddNode(std::unique_ptr<Node> node) {
		node->SetEventCallbackFn([this](Event& e) {
				this->OnEvent(e);
			});
		all_nodes_.push_back(std::move(node));
	}

	void Client::Run() {
		using namespace std::chrono;
		using namespace std::chrono_literals;

		for (auto& node : all_nodes_) {
			StartNode(node.get());
		}

		CLIENT_INFO("Client running.")
		auto curr_time = clock::now();
		auto last_update = curr_time;
		while (running_) {
			curr_time = clock::now();
			nanoseconds delta_time = (curr_time - last_update);
			last_update = curr_time;

			for (const auto& node : active_nodes_) {
				node->OnUpdate(delta_time);
			}
			window_->OnUpdate();
		}

		//Cleanup after ShutDownEvent
		while(!active_nodes_.empty()) {
			StopNode(active_nodes_[0]);
		}

		CLIENT_INFO("All active nodes stopped.")
	}

	void Client::OnEvent(Event& e) {

		EventDispatcher dispatcher(e);
		dispatcher.Dispatch<ShutdownEvent>([this]() -> bool {
			return OnShutdownEvent();
		});
		dispatcher.Dispatch<WindowCloseEvent>([this]() -> bool {
			return OnWindowCloseEvent();
		});
		dispatcher.Dispatch<WindowResizeEvent>([this](const WindowResizeEvent& e) -> bool {
			return OnWindowResizeEvent(e);
		});

		for (const auto& node : active_nodes_) {
			if (e.handled) { break; }
			node->OnEvent(e);
		}
		//Maybe add debug assert on event not handled
	}

	void Client::StartNode(Node* node) {

		node->OnStartup();
		active_nodes_.push_back(node);
	}

	void Client::StopNode(Node* node) {

		auto it = std::find(active_nodes_.begin(), active_nodes_.end(), node);
		if (it == active_nodes_.end()) {
			CLIENT_WARN("Trying to stop inactive node")
			return;
		}
		active_nodes_.erase(it);
		node->OnShutdown();
	}

	bool Client::OnShutdownEvent() {
		running_ = false;
		return false;
	}

	bool Client::OnWindowCloseEvent() {
		running_ = false;
		return false;
	}

	bool Client::OnWindowResizeEvent([[gnu::unused]] const WindowResizeEvent& e) {

		return false;
	}

}} //namespace airsim_client::core
