// Thanks to Yan Chernikov for the Client and Event handling system
// It is based on his work with the Hazel game engine.
// github: https://github.com/TheCherno/Hazel
// youtube: https://www.youtube.com/user/TheChernoProject

#ifndef AIRSIM_CLIENT_EVENT
#define AIRSIM_CLIENT_EVENT
#define BIT(x) (1 << (x))

#include "Log.hpp"

#include <string>
#include <functional>
#include <iostream>
#include <sstream>

namespace airsim_client {

	// EventType is defined so that the EventDispatcher can choose the correct
	// Event handling function to handle the Event.
	// Add new types if needed.
	enum class EventType {

		None = 0,
		Shutdown, WindowClose, WindowResize,
		KeyPressed, KeyReleased,
		DataTransfer,

		// User defined Types
		Image, Transform, Imu,
		TransformPair
	};

	// EventCategory is used if the user wishes to filter out Categories of Events
	// For example for logging purposes.
	// Add new categories if needed.
	enum EventCategory {
		EventCategoryCore = 0,
		EventCategoryClient = BIT(1),
		EventCategoryKeyboard = BIT(2),
		EventCategoryDataTransfer = BIT(3),

		// User defined categories
		EventCategoryAirSim = BIT(4),
		EventCategoryRos = BIT(5)
	};

	// Add the EVENT_CLASS_TYPE macro to make the Event the disignated EventType.
	// It also defines some functions used by the EventDispatcher.
	// The Macro must be added as public in the class definition.
	#define EVENT_CLASS_TYPE(type) static constexpr EventType GetStaticType() { return type; }\
									EventType GetEventType() const override { return GetStaticType(); }\
									const char* GetName() const override { return #type; }

	// Add the EVENT_CLASS_CATEGORY macro to add search categories.
	// The Macro should only be called once. If multiple categories is wanted
	// use bitwise or. Example:
	// EVENT_CLASS_CATEGORY(EventCategoryKeyBoard | EventCategoryClient)
	// The macro must be added as public in the class definition.
	#define EVENT_CLASS_CATEGORY(category) int GetCategoryFlags() const override { return category; }

	} // namespace airsim_client

	//*******************************************//
	//             Event classes                 //
	//*******************************************//

	namespace airsim_client { namespace core {

	class Event {

	public:
		bool handled = false;

		virtual EventType GetEventType() const = 0;
		virtual const char* GetName() const = 0; // might not need this
		virtual int GetCategoryFlags() const = 0;
		virtual std::string ToString() const { return GetName(); } // For debugging only

		inline bool IsInCategory(EventCategory category) {
			return static_cast<bool>(GetCategoryFlags() & category);
		}
	};

	// EventDispatcher should be defined in the OnEvent function of
	// the Node which should handle a specific Event. It is used
	// to bind event handling functions to an EventType and call
	// that function when an Event of that type is created.

	// Usage (in NodeName.cpp):
	// void NodeName::OnEvent(Event& event)
	// {
	// 		EventDispatcher dispatcher(event);
	//		dispatcher.Dispatch<EventType>(bind your event handling function);
	//		dispatcher.Dispatch<ExampleEventType>([this](const ExampleEventType& event) {
	//			-- Do something --
	//		});
	//		dispatcher.Dispatch<EventType2>(---)
	//		--- add as many as you want ---
	// }
	class EventDispatcher {

		template<typename T>
		using EventFunction = std::function<bool(const T&)>;

		using EventFnNoArg = std::function<bool()>;

	public:
		explicit EventDispatcher(Event& event)
			: event_{event} {}

		template<typename T>
		bool Dispatch(EventFunction<T> func) {
			if (event_.GetEventType() == T::GetStaticType()) {
				CLIENT_EVENT(event_.ToString())
				event_.handled = func(*(T*)&event_);
				return true;
			}
			return false;
		}

		template<typename T>
		bool Dispatch(EventFnNoArg func) {
			if (event_.GetEventType() == T::GetStaticType()) {
				CLIENT_EVENT(event_.ToString())
				event_.handled = func();
				return true;
			}
			return false;
		}

	private:
		Event& event_;
	};

	inline std::ostream& operator << (std::ostream& os, const Event& e) {
		return os << e.ToString();
	}

}} //namespace airsim_client::core

#endif //AIRSIM_CLIENT_EVENT
