#ifndef AIRSIM_CLIENT_NODE
#define AIRSIM_CLIENT_NODE

#include "Core/Log.hpp"
#include <functional>

namespace airsim_client { namespace core {

	class Event;

	class Node {
	public:
		Node() = default;
		virtual ~Node() = default;
		Node(const Node& node) = default;
		Node(Node&& node) noexcept = default;
		Node& operator = (const Node& other) = default;
		Node& operator = (Node&& other) noexcept = default;

	protected:
		// Convenience typedefs to use in Classes which inherits from Node
		template <typename Ratio>
		using Duration = std::chrono::duration<float, Ratio>;
		using Seconds = Duration<std::ratio<1>>;
		using Milliseconds = Duration<std::milli>;
		using Microseconds = Duration<std::micro>;
		using Nanoseconds = Duration<std::nano>;

		using EventCallbackFn = std::function<void (Event&)>;

	public:
		// OnStartup is called automaticly by the Client::StartNode function,
		// which is called before the Client's run loop.
		// Use this to set up stuff that needs to be set up before the update loop starts,
		// and that needs access to Client Events or Logging.
		virtual void OnStartup() = 0;

		// OnShutdown is called by Client::StopNode automaticly
		// after after a ShutdownEvent, before the Node destructor.
		// Use it if you need to create events to tell other nodes that is it stopping,
		// or call shutdown methods that might throw exceptions.
		virtual void OnShutdown() = 0;

		// OnUpdate is called automaticly each Client update loop.
		// Use this to perform regularilly update tasks in the node.
		// This function should not stall the update loop.
		// Use async operations if needed.
		virtual void OnUpdate(Seconds dt) = 0;

		// Is called by the Client when it receives an Event.
		// The Client will loop through and call the active Nodes'OnEvent functions
		// until the Event is marked as handled.
		// Use this function to dispatch Event handling functions, using the
		// EventDispatcher's Dispatch function defined in Event.hpp
		virtual void OnEvent(Event& e) = 0;

		// Inheriting classes should define a private variable of type EventCallbackFn.
		// The Client will call SetEventCallbackFn during Client::StartNode to Enable
		// Events for the Node.
		// When generating Events. This function should be called with the Event as an
		// argument.
		virtual void SetEventCallbackFn(const EventCallbackFn& callback) = 0;
	};

}} //namespace airsim_client::core
#endif // AIRSIM_CLIENT_NODE
