#include "Window.hpp"
#include "Log.hpp"
#include "Events/KeyEvent.hpp"
#include "Events/ClientEvent.hpp"

#include <GLFW/glfw3.h>

namespace airsim_client { namespace core {

	static void GLFWErrorCallback(int error, const char* description)
	{
		CLIENT_ERROR("GLFW Error ({0}): {1}", error, description);
	}

	Window::Window(const WindowProperties& props) 
		: data_{props}
	{

		CLIENT_INFO("Creating window {0} ({1}, {2})", props.title, props.width, props.height);
		
		if (!glfwInit()) {
			CLIENT_ERROR("Could not initialize GLFW!")
			return;
		}

		glfwSetErrorCallback(GLFWErrorCallback);

		glfwWindowHint(GLFW_RESIZABLE, GLFW_TRUE);
		window_ = glfwCreateWindow(props.width, props.height, props.title.c_str(), nullptr, nullptr);
		glfwMakeContextCurrent(window_);
		glfwSetWindowUserPointer(window_, &data_);
		glfwSwapInterval(0);


		glfwSetKeyCallback(window_, [](GLFWwindow* window, int key, [[gnu::unused]] int scancode, int action, [[gnu::unused]] int mods)
			{
				WindowData& data = *(WindowData*)glfwGetWindowUserPointer(window);

				if (key == GLFW_KEY_ESCAPE) {
					WindowCloseEvent event;
					data.EventCallback(event);
					return;
				}

				bool repeated = false;
				switch (action) 
				{
					case GLFW_REPEAT:
						repeated = true;
					case GLFW_PRESS: {
						KeyPressedEvent event(key, repeated);
						data.EventCallback(event);
						break;
					}

					case GLFW_RELEASE: {
						KeyReleasedEvent event(key);
						data.EventCallback(event);
						break;
					}
				}
			});

		glfwSetWindowSizeCallback(window_, [](GLFWwindow* window, int width, int height)
		{
			WindowData& data = *(WindowData*)glfwGetWindowUserPointer(window);
			data.width = width;
			data.height = height;
			WindowResizeEvent event(width, height);
			data.EventCallback(event);
		});

		glfwSetWindowCloseCallback(window_, [](GLFWwindow* window)
		{
			WindowData& data = *(WindowData*)glfwGetWindowUserPointer(window);
			WindowCloseEvent event;
			data.EventCallback(event);
		});

	}

	Window::~Window() noexcept {
		glfwDestroyWindow(window_);
		glfwTerminate();
	}

	void Window::OnUpdate() {
		glfwPollEvents();
		glfwSwapBuffers(window_);
	}

}} //namespace airsim_client::core
