#ifndef AIRSIM_CLIENT_AIRSIM_EVENT
#define AIRSIM_CLIENT_AIRSIM_EVENT

#include "Core/Events/DataTransferEvent.hpp"
#include "AirSimTools/Vehicle.hpp"
#include "AirSimTools/CommonTypes.hpp"

#include <common/ImageCaptureBase.hpp>
#include <physics/Kinematics.hpp>
#include <sensors/SensorIncludes.hpp>
#include <utility>
#include <vector>

namespace airsim_client {

	class ImageEvent : public core::DataTransferEvent<std::vector<CameraInfo>, std::vector<Image>> {

	public:
		EVENT_CLASS_CATEGORY(EventCategoryAirSim | EventCategoryDataTransfer)
		EVENT_CLASS_TYPE(EventType::Image)

		ImageEvent(std::string vehicle_name, std::vector<CameraInfo> infos, std::vector<Image> images)
			: DataTransferEvent{std::move(infos), std::move(images)}, vehicle_name_{std::move(vehicle_name)} {}

		inline auto GetVehicleName() const -> std::string { return vehicle_name_; }
		inline auto GetImages() const ->
			std::pair<std::shared_ptr<std::vector<CameraInfo>>, std::shared_ptr<std::vector<Image>>>
		{
			return make_pair(std::get<0>(data_), std::get<1>(data_));
		}

	private:
		std::string vehicle_name_;
	};


	class TransformEvent : public core::DataTransferEvent<std::vector<Transform>> {

	public:
		EVENT_CLASS_CATEGORY(EventCategoryAirSim | EventCategoryDataTransfer)
		EVENT_CLASS_TYPE(EventType::Transform)

		TransformEvent(std::vector<Transform> tfs, bool is_static)
			: DataTransferEvent{std::move(tfs)}, is_static_{is_static} {}

		constexpr bool IsStatic() const { return is_static_; }
		inline auto GetTransforms() const -> std::vector<Transform> { return *std::get<0>(data_); }

		std::string ToString() const override
		{
			std::stringstream ss;

			ss 	<< "TransformEvent:\n"
				<< "\tStaticEvent? " << is_static_ << "\n";

			return ss.str();
		}

	private:
		bool is_static_;
	};

	template <typename SensorType> class SensorEvent;
	using Imu = msr::airlib::ImuBase::Output;
	using ImuEvent = SensorEvent<Imu>;


	// SensorEvent Implementations

	template <>
	class SensorEvent<Imu> : public core::DataTransferEvent<Imu> {
	public:
		EVENT_CLASS_CATEGORY(EventCategoryAirSim | EventCategoryDataTransfer)
		EVENT_CLASS_TYPE(EventType::Imu)

		SensorEvent(std::string vehicle_name, Imu data)
			: DataTransferEvent{std::move(data)}, vehicle_name_{std::move(vehicle_name)} {}

		inline auto GetSensorData() const -> std::pair<std::string, Imu>
		{
			return make_pair(vehicle_name_, *std::get<0>(data_));
		}

	private:
		std::string vehicle_name_;
	};

} //namespace airsim_client
#endif //AIRSIM_CLIENT_AIRSIM_EVENT
