#include "CvNode.hpp"
#include "Core/Events/ClientEvent.hpp"
#include "AirSimTools/SettingsParser.hpp"

#include <rpc/rpc_error.h>

namespace airsim_client {

	CvNode::CvNode() 
		: Client{std::make_unique<RpcClient>()}
	{
		CLIENT_INFO("CvNode Created")
	}

	CvNode::~CvNode() noexcept
	{
		CLIENT_INFO("CvNode Destroyed")
	}

	void CvNode::OnStartup() {

		try {
			initializeVehicleSettings();

			Client->waitForConnection();
		}
		catch (std::runtime_error& e) {
			CLIENT_ERROR("Caught error {0} in CvNode", e.what())
			core::ShutdownEvent event;
			EventCallback(event);
			return;
		}

		requests_.emplace_back("front_center", msr::airlib::ImageCaptureBase::ImageType::Scene, false, false);
	}

	void CvNode::OnShutdown() {
		try {
			Client->disconnect();
		}
		catch (rpc::timeout& e) {
			CLIENT_ERROR(e.what())
		}
		catch (rpc::rpc_error& e) {
			CLIENT_ERROR(e.get_error().as<std::string>())
		}
	}

	void CvNode::OnUpdate([[gnu::unused]] Seconds dt) {
		try {
			const auto& camera_infos = vehicle_->getCameraInfos();
			const auto& response = Client->simGetImages(requests_, vehicle_->getName());

			std::vector<Image> images;
			images.reserve(response.size());
			std::vector<CameraInfo> relevant_info;
			relevant_info.reserve(response.size());

			for (const auto& image : response) {
				images.emplace_back(image);
				auto info = camera_infos.at(SettingsParser::getCaptureType(image.image_type));
				info.time_stamp = TimeStamp{std::chrono::duration<double, std::nano>{image.time_stamp}};
				relevant_info.push_back(info);
			}
			ImageEvent imageEvent(vehicle_->getName(), std::move(relevant_info), std::move(images));
			EventCallback(imageEvent);
		}
		catch (rpc::timeout& e) {
			CLIENT_ERROR(e.what())
			core::ShutdownEvent event;
			EventCallback(event);
		}
		catch (rpc::rpc_error& e) {
			CLIENT_ERROR(e.get_error().as<std::string>())
			core::ShutdownEvent event;
			EventCallback(event);
		}
	}

	void CvNode::OnEvent([[gnu::unused]] core::Event& e) {
		//EventDispatcher dispatcher(e);
	}

	void CvNode::initializeVehicleSettings() {

		for (const auto& [name, setting] : SettingsParser::Get().VehicleMap()) {
			vehicle_ = std::make_unique<Vehicle>("world_ned", *setting);
		}
	}

} //namespace airsim_client
