#include "MultirotorNode.hpp"
#include "Core/Input.hpp"
#include "Core/Events/ClientEvent.hpp"
#include "AirSimTools/SettingsParser.hpp"

#include <common/AirSimSettings.hpp>

#include <rpc/rpc_error.h>
#include <GLFW/glfw3.h>

namespace airsim_client {

	using RpcClient = msr::airlib::MultirotorRpcLibClient;

	MultirotorNode::MultirotorNode(Milliseconds update_interval)
		: Client{std::make_unique<RpcClient>("localhost", 41451, 5)}, update_interval_{update_interval}
	{
		CLIENT_INFO("Creating MultirotorNode")
	}

	MultirotorNode::~MultirotorNode() noexcept
	{
		CLIENT_INFO("Destroying MultirotorNode")
	}

	void MultirotorNode::OnStartup() {

		try {
			initializeVehicleSettings();

			Client->confirmConnection();
			if (!isConnected())
			{
				CLIENT_ERROR("Could not connect to the RPC server")
				core::ShutdownEvent event;
				EventCallback(event);
				return;
			}

			Client->enableApiControl(true);
			Client->armDisarm(true);
		}

		catch (std::runtime_error& e) {
			CLIENT_ERROR("Could not initialize Vehicles")
			core::ShutdownEvent event;
			EventCallback(event);
			return;
		}

		requests_.emplace_back("front_center", msr::airlib::ImageCaptureBase::ImageType::Scene, false, false);
		requests_.emplace_back("front_center", msr::airlib::ImageCaptureBase::ImageType::DepthPlanner, true);
	}

	void MultirotorNode::OnShutdown() {
		if (isConnected()) {
			try {
				Client->waitOnLastTask();
				Client->armDisarm(false);
				Client->enableApiControl(false);
			}
			catch (rpc::timeout& e) {
				CLIENT_ERROR(e.what())
			}
			catch (rpc::rpc_error& e) {
				CLIENT_ERROR(e.get_error().as<std::string>())
			}
		}
	}

	void MultirotorNode::OnUpdate(Seconds dt) {

		if (!isConnected()) {
			CLIENT_WARN("MultirotorNode not connected to RPC server")
			return;
		}

		if (static_tf_msgs_updated_) {
			TransformEvent transformEvent(static_tf_msgs_, true);
			EventCallback(transformEvent);
			static_tf_msgs_updated_ = false;
		}

		for (auto& multirotor : multirotors_) {
			try {
				updateAndSendTransform(multirotor);
				setVelocity(multirotor);
				accumulated_time_ += dt;
				if (accumulated_time_ >= update_interval_)
				{
					accumulated_time_ -= update_interval_;
					while (accumulated_time_ > update_interval_) {
						CLIENT_WARN("Frame missed in MultirotorNode::OnUpdate")
						accumulated_time_ -= update_interval_;
					}
					receiveAndSendImages(multirotor);
					receiveAndSendSensorData(multirotor);
				}
			}

			catch(std::runtime_error& e) {
				CLIENT_ERROR("MultirotorNode runtime error: {0}", e.what())
				core::ShutdownEvent event;
				EventCallback(event);
				return;
			}
		}
	}

	void MultirotorNode::OnEvent([[gnu::unused]] core::Event& e) {}

	void MultirotorNode::initializeVehicleSettings() {

		for (auto& vehicle_setting : SettingsParser::Get().VehicleMap()) {
			multirotors_.emplace_back("world_ned", *vehicle_setting.second);
		}

		for (auto& multirotor : multirotors_) {
			std::vector<Transform> vec = multirotor.getStaticTransforms();
			static_tf_msgs_.reserve(static_tf_msgs_.size() + vec.size());
			static_tf_msgs_.insert(std::end(static_tf_msgs_), std::begin(vec), std::end(vec));
		}
		static_tf_msgs_updated_ = true;
	}

	void MultirotorNode::updateAndSendTransform(Vehicle& vehicle) const {
		using namespace msr::airlib;

		MultirotorState state = Client->getMultirotorState(vehicle.getName());
		Kinematics::State kinematics = Client->simGetGroundTruthKinematics(vehicle.getName());
		vehicle.updateCurrentTransform(kinematics, TimeStamp{std::chrono::duration<double, std::nano>{state.timestamp}});

		TransformEvent transform_event({vehicle.getCurrentTransform()}, false);
		EventCallback(transform_event);
	}

	void MultirotorNode::setVelocity(Vehicle& vehicle) const {
		using Input = core::Input;
		Input::IsKeyPressed(GLFW_KEY_W) 	? 	vehicle.vel_body_.x() = 1.0f 	:
		Input::IsKeyPressed(GLFW_KEY_S) 	? 	vehicle.vel_body_.x() = -1.0f 	:
												vehicle.vel_body_.x() = 0.0f;

		Input::IsKeyPressed(GLFW_KEY_D) 	? 	vehicle.vel_body_.y() = 1.0f 	:
		Input::IsKeyPressed(GLFW_KEY_A) 	? 	vehicle.vel_body_.y() = -1.0f 	:
												vehicle.vel_body_.y() = 0.0f;

		Input::IsKeyPressed(GLFW_KEY_SPACE) ? 	vehicle.vel_body_.z() = -1.0f 	:
		Input::IsKeyPressed(GLFW_KEY_X) 	? 	vehicle.vel_body_.z() = 1.0f 	:
												vehicle.vel_body_.z() = 0.0f;


		vehicle.vel_body_.normalize();
		vehicle.vel_body_ = vehicle.vel_body_ * vehicle.speed_;

		using namespace msr::airlib;

		// Not exact, but good enough.
		Pose world_pose = vehicle.getCurrentTransform().transform;
		VectorMath::Quaternionf yaw_rotation = world_pose.orientation;
		yaw_rotation.x() = 0; yaw_rotation.y() = 0;
		yaw_rotation.normalize();

		VectorMathf::Quaternionf q;
		q.w() = 0;
		q.vec() = vehicle.vel_body_;
		q = yaw_rotation * q * yaw_rotation.inverse();
		vehicle.vel_world_ = q.vec();

		Client->moveByVelocityAsync(
			vehicle.vel_world_.x(),
			vehicle.vel_world_.y(),
			vehicle.vel_world_.z(),
			0.05,
			msr::airlib::DrivetrainType::ForwardOnly,
			msr::airlib::YawMode{false, 0.0f},
			vehicle.getName()
		);
	}

	void MultirotorNode::receiveAndSendImages(const Vehicle& vehicle) const {
		const auto& response = Client->simGetImages(requests_, vehicle.getName());
		const auto& camera_infos = vehicle.getCameraInfos();

		std::vector<Image> images;
		images.reserve(response.size());
		std::vector<CameraInfo> relevant_info;
		relevant_info.reserve(response.size());
		for (const auto& image : response) {
			images.emplace_back(image);
			auto info = camera_infos.at(SettingsParser::getCaptureType(image.image_type));
			info.time_stamp = TimeStamp{std::chrono::duration<double, std::nano>{image.time_stamp}};
			relevant_info.push_back(std::move(info));
		}
		ImageEvent image_event{vehicle.getName(), std::move(relevant_info), std::move(images)};
		EventCallback(image_event);
	}

	void MultirotorNode::receiveAndSendSensorData(const Vehicle& vehicle) const {
		ImuEvent imuEvent(vehicle.getName(), Client->getImuData("", vehicle.getName()));
		EventCallback(imuEvent);
	}






} //namespace airsim_client
