#include "RosNode.hpp"

#include "Core/Events/ClientEvent.hpp"
#include "AirSimTools/SettingsParser.hpp"

#include <std_msgs/String.h>
#include <sensor_msgs/CameraInfo.h>
#include <sensor_msgs/Image.h>
#include <sensor_msgs/Imu.h>
#include <opencv2/opencv.hpp>
#include <cv_bridge/cv_bridge.h>

#include <thread>

namespace airsim_client {

	RosNode::RosNode(int argc, char** argv) {
		ros::init(argc, argv, "airsim_client");
		nh_ = std::make_unique<ros::NodeHandle>();
		nh_private_ = std::make_unique<ros::NodeHandle>("~");
		nh_private_->param<bool>("param_with_slam", param_with_slam, false);
		nh_private_->param<bool>("param_greyscale", param_greyscale, false);
		CLIENT_INFO("Creating RosNode")
	}

	RosNode::~RosNode() noexcept {
		ros::shutdown();
		CLIENT_INFO("Destroying RosNode")
	}

	void RosNode::OnStartup() {

		tf_listener_ = std::make_unique<tf2_ros::TransformListener>(tf2_buffer_);

		try {
			initializeVehiclePublishers();
		}
		catch (std::runtime_error& e) {
			CLIENT_ERROR("Caught error: {0} in RosNode", e.what())
			core::ShutdownEvent event;
			EventCallback(event);
			return;
		}

		tf_publisher_ = std::make_unique<tf2_ros::TransformBroadcaster>();
		tf_static_publisher_ = std::make_unique<tf2_ros::StaticTransformBroadcaster>();
	}

	void RosNode::OnShutdown() {
	}

	void RosNode::OnUpdate([[gnu::unused]] Seconds dt) {

		if (!ros::ok()) {
			core::ShutdownEvent event;
			EventCallback(event);
			return;
		}

		for (auto& msg : static_tf_msgs_) {
			msg.header.stamp = ros::Time::now();
			tf_static_publisher_->sendTransform(msg);
		}

		if (param_with_slam) { matchTransforms(); }

		ros::spinOnce();
	}

	void RosNode::OnEvent(core::Event& e) {
		core::EventDispatcher dispatcher(e);
		dispatcher.Dispatch<ImageEvent>([this](const ImageEvent& e) -> bool {
			return OnImageEvent(e);
		});
		dispatcher.Dispatch<ImuEvent>([this](const ImuEvent& e) -> bool {
			return OnImuEvent(e);
		});
		dispatcher.Dispatch<TransformEvent>([this](const TransformEvent& e) -> bool {
			return OnTransformEvent(e);
		});
	}

	bool RosNode::OnImageEvent(const ImageEvent& e) {
		using ImageType = msr::airlib::ImageCaptureBase::ImageType;

		const auto [infos, images] = e.GetImages();

		for (unsigned int i = 0; i < images->size(); i++) {
			auto image = images->at(i);

			if (image.image_type == ImageType::Scene) {
				publishSceneImage(image, infos->at(i), e.GetVehicleName());
			}
			else if (image.image_type == ImageType::DepthPlanner) {
				publishDepthImage(image, infos->at(i), e.GetVehicleName());
			}
			else {
				CLIENT_ERROR("Could not publish image. Unsupported type")
			}
		}

		return true;
	}

	bool RosNode::OnImuEvent(const ImuEvent& e) {

		const auto [vehicle_name, imuData] = e.GetSensorData();

		geometry_msgs::Quaternion orientation;
		orientation.x = imuData.orientation.x();
		orientation.y = imuData.orientation.y();
		orientation.z = imuData.orientation.z();
		orientation.w = imuData.orientation.w();

		geometry_msgs::Vector3 ang_vel;
		ang_vel.x = imuData.angular_velocity(0);
		ang_vel.y = imuData.angular_velocity(1);
		ang_vel.z = imuData.angular_velocity(2);

		geometry_msgs::Vector3 lin_acc;
		lin_acc.x = imuData.linear_acceleration(0);
		lin_acc.y = imuData.linear_acceleration(1);
		lin_acc.z = imuData.linear_acceleration(2);

		sensor_msgs::Imu imuMessage;
		imuMessage.header.stamp = ros::Time{static_cast<double>(imuData.time_stamp)/1E9};
		imuMessage.orientation = orientation;
		imuMessage.angular_velocity = ang_vel;
		imuMessage.linear_acceleration = lin_acc;

		vehicle_publishers_[vehicle_name].imu_publisher_.publish(imuMessage);

		return true;
	}

	bool RosNode::OnTransformEvent(const TransformEvent& e) {

		for (auto& transform : e.GetTransforms()) {
			geometry_msgs::TransformStamped tf;
			tf.header.stamp = ros::Time{transform.time_stamp.time_since_epoch().count()};
			tf.header.frame_id = transform.parent_frame_id;
			tf.child_frame_id = transform.child_frame_id;

			tf.transform.rotation.x = transform.transform.orientation.x();
			tf.transform.rotation.y = transform.transform.orientation.y();
			tf.transform.rotation.z = transform.transform.orientation.z();
			tf.transform.rotation.w = transform.transform.orientation.w();
			tf.transform.translation.x = transform.transform.position.x();
			tf.transform.translation.y = transform.transform.position.y();
			tf.transform.translation.z = transform.transform.position.z();

			if (e.IsStatic()) {
				static_tf_msgs_.push_back(tf);
			}
			else {
				tf_publisher_->sendTransform(tf);
				clock::time_point curr = clock::now();
				if ((curr - last_tf_sample_) > std::chrono::milliseconds{500}) {
					unpaired_tfs_.Add(std::move(tf));
					last_tf_sample_ = curr;
				}
			}
		}

		return true;
	}

	sensor_msgs::CameraInfo RosNode::makeCameraInfoMsg(const CameraInfo& info) const {

		sensor_msgs::CameraInfo msg;
		msg.header.frame_id = info.optical_frame_id;
		msg.header.stamp = ros::Time{info.time_stamp.time_since_epoch().count()};
		msg.width = info.image_width;
		msg.height = info.image_height;

		auto& K = info.intrinsic_matrix;

		msg.K = {
			K(0,0), K(0,1), K(0,2),
			K(1,0), K(1,1), K(1,2),
			K(2,0), K(2,1), K(2,2)
		};

		msg.P = {
			K(0,0), K(0,1), K(0,2), info.position.x(),
			K(1,0), K(1,1), K(1,2), info.position.y(),
			K(2,0), K(2,1), K(2,2), info.position.z()
		};

		return msg;
	}

	void RosNode::publishSceneImage(Image& image, const CameraInfo& info, const std::string& vehicle_name) const {
		cv::Mat mat(image.height, image.width, CV_8UC4, image.image_data_uint8.data());
		if (!mat.empty()) {
			std_msgs::Header header;
			header.frame_id = info.optical_frame_id;
			header.stamp = ros::Time{info.time_stamp.time_since_epoch().count()};;

			cv_bridge::CvImage cv_img;
			cv_img.header = header;

			if (param_greyscale) {
				cv::cvtColor(mat, mat, cv::COLOR_RGBA2GRAY);
				cv_img.image = mat;
				cv_img.encoding = "mono8";
			}
			else {
				cv::cvtColor(mat, mat, cv::COLOR_RGBA2RGB); //RGBA8 not supported by Kalibr toolbox
				cv_img.image = mat;
				cv_img.encoding = "rgb8";
			}

			sensor_msgs::Image ros_img;
			cv_img.toImageMsg(ros_img);

			sensor_msgs::CameraInfo info_msg = makeCameraInfoMsg(info);

			vehicle_publishers_.at(vehicle_name).camera_info_publishers_.at(SettingsParser::getCaptureType(image.image_type)).publish(info_msg);
			vehicle_publishers_.at(vehicle_name).camera_image_publishers_.at(SettingsParser::getCaptureType(image.image_type)).publish(ros_img);
		}
	}

	void RosNode::publishDepthImage(Image& image, const CameraInfo& info, const std::string& vehicle_name) const {

		cv::Mat depth_img(image.height, image.width, CV_32FC1, image.image_data_float.data());

		if(!depth_img.empty()) {
			std_msgs::Header header;
			header.frame_id = info.optical_frame_id;
			header.stamp = ros::Time{info.time_stamp.time_since_epoch().count()};
			cv_bridge::CvImage cv_img(header, "32FC1", depth_img);

			sensor_msgs::Image ros_img;
			cv_img.toImageMsg(ros_img);

			sensor_msgs::CameraInfo info_msg = makeCameraInfoMsg(info);

			vehicle_publishers_.at(vehicle_name).camera_info_publishers_.at(SettingsParser::getCaptureType(image.image_type)).publish(info_msg);
			vehicle_publishers_.at(vehicle_name).camera_image_publishers_.at(SettingsParser::getCaptureType(image.image_type)).publish(ros_img);
		}
	}

	// Only for one camera for now. It is a future todo
	void RosNode::initializeVehiclePublishers() {

		for (auto& vehicle : SettingsParser::Get().VehicleMap()) {
			vehicle_publishers_[vehicle.first] = VehiclePublishers();
			auto& vehicle_setting = *vehicle.second;

			vehicle_publishers_[vehicle.first].imu_publisher_ = nh_->advertise<sensor_msgs::Imu>("/" + vehicle.first + "/sensor/imu", 2);

			for (auto& camera : vehicle_setting.cameras) {
				for (auto& capture_setting : camera.second.capture_settings) {
					if (!std::isnan(capture_setting.second.fov_degrees)) {

						std::string base_topic = vehicle.first + "/" + camera.first + "/" + SettingsParser::getCaptureTypeName(capture_setting.first);

						vehicle_publishers_[vehicle.first].camera_info_publishers_[capture_setting.first] =
							nh_->advertise<sensor_msgs::CameraInfo>(base_topic + "/info", 2);
						vehicle_publishers_[vehicle.first].camera_image_publishers_[capture_setting.first] =
							nh_->advertise<sensor_msgs::Image>(base_topic + "/image_raw", 2);
					}
				}
			}

		}
	}

	void RosNode::matchTransforms() {
		auto tf_vec = unpaired_tfs_.Dump();
		try {
			geometry_msgs::TransformStamped new_tf = tf2_buffer_.lookupTransform("camera_voslam", "world_nwu", ros::Time{0});
			for (auto& tf : tf_vec) {
				if (std::abs(tf.header.stamp.toSec() - new_tf.header.stamp.toSec()) < 0.05) {
					//CLIENT_INFO("Match!")
					GroundTruthEstimatePairEvent<TF, TF> event{std::make_pair(tf, new_tf)};
					EventCallback(event);
				}
				else if (tf.header.stamp > new_tf.header.stamp) {
					//CLIENT_INFO("Future!")
					unpaired_tfs_.Add(tf);
				}
			}
		}
		catch (tf2::TransformException& e) {
			CLIENT_WARN("Could not get Transform: {0}", e.what())
		}
	}
} //namespace airsim_client
