#ifndef AIRSIM_CLIENT_ROS_NODE
#define AIRSIM_CLIENT_ROS_NODE

#include "Core/Node.hpp"
#include "Events/RosEvent.hpp"
#include "Events/AirSimEvent.hpp"

#include <ros/ros.h>
#include <image_transport/image_transport.h>
#include <tf2_geometry_msgs/tf2_geometry_msgs.h>
#include <tf2_ros/transform_broadcaster.h>
#include <tf2_ros/static_transform_broadcaster.h>
#include <tf2_ros/transform_listener.h>
#include <memory>
#include <vector>

namespace airsim_client {

	class RosNode : public core::Node {
	public:
		struct VehiclePublishers
		{
			std::unordered_map<int, ros::Publisher> camera_info_publishers_;
			std::unordered_map<int, ros::Publisher> camera_image_publishers_;

			ros::Publisher imu_publisher_;
		};

		template <std::size_t N>
		class TransformBuffer
		{
		private:
			int curr_index_ = 0;
			int valid_elems_ = 0;
			std::array<geometry_msgs::TransformStamped, N> arr_;

		public:
			void Add(geometry_msgs::TransformStamped tf)
			{
				arr_.at(curr_index_) = std::move(tf);
				curr_index_ = (++curr_index_)%N;
				valid_elems_ < static_cast<int>(N) ? valid_elems_++ : valid_elems_ = N;
			};
			auto Dump() -> std::vector<geometry_msgs::TransformStamped>
			{
				std::vector<geometry_msgs::TransformStamped> v; v.reserve(valid_elems_);
				for (int i = 0; i < valid_elems_; ++i)
				{
					v.push_back(std::move(arr_.at(i)));
				}
				valid_elems_ = 0; curr_index_ = 0;
				return v;
			}
		};

	public:
		RosNode(int argc, char** argv);
		~RosNode() noexcept final;

		// Should be held and managed by the Client for the whole program duration
		RosNode(const RosNode&) = delete;
		RosNode& operator = (const RosNode&) = delete;
		RosNode(RosNode&&) noexcept = delete;
		RosNode& operator = (RosNode&&) noexcept = delete;

	public:
		void OnStartup() final;
		void OnShutdown() final;
		void OnUpdate(Seconds dt) final;
		void OnEvent(core::Event& e) final;

	inline void SetEventCallbackFn(const EventCallbackFn& callback) final { EventCallback = callback; }

	private:

		bool OnRosNotOkEvent();
		bool OnImageEvent(const ImageEvent& e);
		bool OnImuEvent(const ImuEvent& e);
		bool OnTransformEvent(const TransformEvent& e);

		sensor_msgs::CameraInfo makeCameraInfoMsg(const CameraInfo& info) const;

		void publishSceneImage(Image& image, const CameraInfo& info, const std::string& vehicle_name) const;
		void publishDepthImage(Image& image, const CameraInfo& info, const std::string& vehicle_name) const;

		void initializeVehiclePublishers();

		void matchTransforms();

	private:
		std::vector<geometry_msgs::TransformStamped> static_tf_msgs_;
		TransformBuffer<20> unpaired_tfs_;
		using clock = std::chrono::steady_clock;
		clock::time_point last_tf_sample_ = clock::now();

		std::unique_ptr<ros::NodeHandle> nh_;
		std::unique_ptr<ros::NodeHandle> nh_private_;
		bool param_with_slam = false;
		bool param_greyscale = false;

		//subscribers & buffers
		std::string target_frame_ = "orb_slam_camera";
		tf2_ros::Buffer tf2_buffer_;
		std::unique_ptr<tf2_ros::TransformListener> tf_listener_;

		//publishers
		std::unique_ptr<tf2_ros::TransformBroadcaster> tf_publisher_;
		std::unique_ptr<tf2_ros::StaticTransformBroadcaster> tf_static_publisher_;
		std::unordered_map<std::string, VehiclePublishers> vehicle_publishers_;

		EventCallbackFn EventCallback;

	};

} // namespace airsim_client
#endif //AIRSIM_CLIENT_ROS_NODE
